import numpy as np
from collections import namedtuple
from scipy.optimize import fsolve, leastsq
from osirisl1services.blurreddata import SAO2010, O3BDM
from osirisl1services.readlevel1 import Orbit, open_level1_attitude, open_level1_spectrograph
from osirisl1services.services import get_tangent_point_data_for_spectrograph_dataset
import logging

CENTRAL_WAVELENGTH = 335.0

# If debugging, use faster to load solar and o3 interps
if logging.getLogger().getEffectiveLevel() == logging.DEBUG:  # pragma: no cover
    logging.warning('Using debug versions of the solar and o3 psf interp objects.')
    solar = SAO2010(sigma=np.arange(.36, .52, 0.02))
    o3 = O3BDM(sigma=np.arange(.36, .52, 0.02))
else:
    solar = SAO2010()
    o3 = O3BDM()


# Target function
def model_sig_mu(sig, mu, p, wv):
    """Simple blur and shift model."""

    n = p[0]
    a = p[1]
    low_order_poly = p[2:]
    wave = wv + mu
    # logging.debug(f'n:{n:3.2e} a:{a:3.2e} sig:{sig:3.2e} mu:{mu:3.2e} '
    #               f'poly:{np.array2string(low_order_poly, precision=3, max_line_width=120)}')

    low_order_spectrum = np.polyval(low_order_poly, wave - CENTRAL_WAVELENGTH)
    return low_order_spectrum + np.log(solar((sig, wave)) * np.exp(-np.exp(n + a * wave) * o3((sig, wave))))


class PointSpreadFunctionFit(object):
    """Class for optimizing the PSF parameters for an OSIRIS scan."""

    min_alt = 47000
    max_alt = 55000

    def __init__(self, scanno, model=model_sig_mu):
        self.scanno = scanno
        self.model = model
        self.popt = None

        # Load data from scan
        att = open_level1_attitude(Orbit(scanno // 1000), scanno=scanno, load_pointing=True)
        spectrograph = open_level1_spectrograph(Orbit(scanno // 1000), scanno=self.scanno)
        lat_lon_alt = get_tangent_point_data_for_spectrograph_dataset(spectrograph=spectrograph, attitude=att)
        spectrograph.merge(lat_lon_alt, inplace=True)

        # deal with nans before dimension swap
        if any(np.isnan(spectrograph.altitude.values)):
            spectrograph = spectrograph.dropna(dim='mjd')
            logging.warning(f'NaN altitudes found in scan {scanno}')

        # reindex to ascending altitiude
        self.scan = spectrograph.swap_dims({'pixel': 'wavelength', 'mjd': 'altitude'}).sortby('altitude').data

    def fit(self, sigma=0.4, mu=0.01, poly_n=5) -> namedtuple:
        """ Fits point spread function to data for each spectrum between 47 and 55 KM

            :param float sigma: initial guess for 1st degree polynomial coefficient
            :param float mu:    initial guess for 0th degree polynomial coefficient
            :param int poly_n:  degree of polynomial to fit to data
        """

        psf_results = namedtuple('PSF_results', ['num_spectra', 'sigma', 'sigma_err', 'shift', 'shift_err', 'sum_sq'])

        spectra = self.scan.sel(wavelength=slice(300, 361), altitude=slice(PointSpreadFunctionFit.min_alt,
                                                                           PointSpreadFunctionFit.max_alt))
        logging.info(f"Number of altitudes used in PSF fit: {spectra.sizes['altitude']}")

        # Create the initial guesses for the c, n, and a parameters
        #   c: absolute brightness of the spectra
        #   n: first guess neutral density parameter
        #   a: first guess ozone cross-sectional parameter
        params = np.array([])

        for _, spectrum in spectra.groupby('altitude'):
            spectrum = spectrum.dropna(dim='wavelength')
            rad305 = spectrum.sel(wavelength=305.3, method='nearest')
            w305 = rad305.wavelength
            rad328 = spectrum.sel(wavelength=328.0, method='nearest')
            w328 = rad328.wavelength

            def n0a0(guess):
                n, a = guess
                return [np.log(np.log(c0 / rad305 * solar((.4, w305))) / o3((.4, w305))) - n - w305 * a,
                        np.log(np.log(c0 / rad328 * solar((.4, w328))) / o3((.4, w328))) - n - w328 * a]

            c0 = spectrum.sel(wavelength=355, method='nearest').values / solar((.4, 355))
            (n0, a0), infodict, ier, errmsg = fsolve(n0a0, np.array([20, .08]), full_output=True)
            logging.info(f"Finding the PSF initial n and a guesses took {infodict['nfev']} iterations")
            logging.info(errmsg)
            logging.debug(f'Altitude {spectrum.altitude.values:4.0f}m: n0: {n0:4.3f}, a0: {a0:4.3f}')
            x = spectrum.wavelength - CENTRAL_WAVELENGTH
            y = np.log(spectrum.values) \
                - [self.model(0.4, 0.0, np.r_[[n0, a0], np.zeros(poly_n + 1)], wv) for wv in spectrum.wavelength]
            p = np.polyfit(x, y, poly_n)
            logging.debug(f'PSF initial poly: {np.array2string(p, precision=3, max_line_width=120)}')
            params = np.r_[params, [n0, a0], p]

        # Vector of the parameters to fit
        # It contains all the per image parameters of the problem, and the per scan parameters
        params = np.r_[sigma, mu, params]

        # Cost function of the fit, compare it to the previous example.
        def errfunc(p, wv, os):
            return np.concatenate(
                list(self.model(p[0], p[1], p[(poly_n + 3) * i + 2:(poly_n + 3) * i + poly_n + 5], wv[np.isfinite(im)])
                     - np.log(im[np.isfinite(im)]) for i, im in enumerate(os))).ravel()

        # Pass the array of datasets, there are one set of wavelengths and multiple spectra.
        self.popt, pcov, infodict, errmsg, ier = leastsq(
            errfunc, params,
            args=(spectra.sel(wavelength=slice(306, 361)).wavelength.values,
                  spectra.sel(wavelength=slice(306, 361)).values),
            full_output=True)

        logging.info(f"Finding the PSF took {infodict['nfev']} iterations")
        logging.info(errmsg)
        for i, (_, spectrum) in enumerate(spectra.groupby('altitude')):
            params = self.popt[(poly_n + 3) * i + 2:(poly_n + 3) * i + poly_n + 5]
            logging.debug(f'Altitude {spectrum.altitude.values:4.0f}m: n: {params[0]:4.3f}, a: {params[1]:4.3f}')
            logging.debug(f'PSF final poly: {np.array2string(params[2:], precision=3, max_line_width=120)}')

        # Calculate the s_sq_err of the residuals
        s_sq_err = np.array(infodict['fvec'] ** 2).sum()
        if ier not in [1, 2, 3, 4]:
            raise RuntimeError("Optimal parameters not found: " + errmsg)

        if pcov is None:
            # indeterminate covariance
            pcov = np.zeros((len(self.popt), len(self.popt)), dtype=float)
            pcov.fill(np.inf)
            # warn_cov = True
            s_sq = s_sq_err / (spectra[:, 80:].size - params.size)
            pcov = pcov * s_sq
        else:
            if spectra[:, 80:].size > params.size:
                s_sq = s_sq_err / (spectra[:, 80:].size - params.size)
                pcov = pcov * s_sq
            else:
                pcov.fill(np.inf)
                raise RuntimeError('Covariance of the parameters could not be estimated')

        perr = np.sqrt(np.diag(pcov))
        logging.info(f'PSF details for {self.scanno}: s_sq={s_sq:4.3e}, '
                     f'sigma={self.popt[0]:4.3f}±{perr[0]:4.3f}, '
                     f'mu={self.popt[1]:5.4f}±{perr[1]:5.4f}')

        return psf_results(len(spectra), self.popt[0], perr[0], self.popt[1], perr[1], s_sq)

    def plot(self):  # pragma: no cover
        """ Plots modelled psf fitting with spectrum data """
        import matplotlib.pyplot as plt
        spectra = self.scan.sel(wavelength=slice(306, 361), altitude=slice(PointSpreadFunctionFit.min_alt,
                                                                           PointSpreadFunctionFit.max_alt))
        poly_n = (len(self.popt) - 2) // spectra.sizes['altitude'] - 3
        sigma = self.popt[0]
        mu = self.popt[1]
        fwhm = sigma * np.sqrt(8 * np.log(2))

        for i, (alt, spectrum) in enumerate(spectra.groupby('altitude')):
            params = self.popt[(poly_n + 3) * i + 2:(poly_n + 3) * i + poly_n + 5]
            modelled = [self.model(sigma, mu, params, wv) for wv in spectrum.wavelength]
            modelled_blur = [self.model(sigma * 1.0, mu-0.05, params, wv) for wv in spectrum.wavelength]
            modelled_sharp = [self.model(sigma / 1.0, mu+0.05, params, wv) for wv in spectrum.wavelength]
            fig, axs = plt.subplots(2, 1, figsize=(10, 10))
            np.log(spectrum).plot.line('k.', ax=axs[0])
            axs[0].plot(spectrum.wavelength, modelled,
                        spectrum.wavelength, modelled_blur,
                        spectrum.wavelength, modelled_sharp)
            axs[0].set_title(f'Scan: {self.scanno}, Polynomial Order: {poly_n}, FWHM: {fwhm:.3}')
            rsq = ((np.log(spectrum) - modelled) ** 2).sum().values.take(0)
            rsq_blur = ((np.log(spectrum) - modelled_blur) ** 2).sum().values.take(0)
            rsq_sharp = ((np.log(spectrum) - modelled_sharp) ** 2).sum().values.take(0)
            axs[1].plot(spectrum.wavelength, np.log(spectrum) - modelled,
                        spectrum.wavelength, np.log(spectrum) - modelled_blur,
                        spectrum.wavelength, np.log(spectrum) - modelled_sharp)
            axs[1].axhline(0, color='k', linestyle='--')
            axs[1].set_title(f'Residual Rsq: {rsq:.4}, Blurred Rsq: {rsq_blur:.4}, Sharp Rsq: {rsq_sharp:.4}, ')
            plt.ylim(-0.06, 0.06)
            plt.show()
